/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "screen_connector_launcher_service/launcher_service.h"
#include "screen_connector_launcher_service/launcher_service_implementation.h"
#include "screen_connector_launcher_service/log_internal.h"

namespace screen_connector {

LauncherService::Impl::Impl(LauncherService* parent,
                            std::shared_ptr<Wl2Window> wl2_win,
                            LauncherService::IEventListener* listener)
  : parent_(parent),
    wl2_win_(std::move(wl2_win)),
    listener_(listener),
    trs_mgr_(new TizenRemoteSurfaceManager()),
    tzsh_(new Tzsh(tzsh_create(TZSH_TOOLKIT_TYPE_EFL))) {
  auto win_id = ecore_wl2_window_id_get(wl2_win_->GetRaw());
  tls_ = std::unique_ptr<TzshLauncherService>(new TzshLauncherService(
        tzsh_launcher_service_create(tzsh_->GetRaw(),
            static_cast<tzsh_window>(win_id))));
  tls_->SetListener(this);

  auto surface = ecore_wl2_window_surface_get(wl2_win_->GetRaw());
  wl_surface_ = std::shared_ptr<WlSurface>(new WlSurface(surface));
}

int LauncherService::Impl::Launch(const std::string& app_id,
                                  const std::string& inst_id,
                                  int pid,
                                  uint32_t serial) {
  return tls_->Launch(app_id, inst_id, pid, serial);
}

int LauncherService::Impl::LaunchWithSharedWidget(const std::string& app_id,
                                                  const std::string& inst_id,
                                                  int pid,
                                                  uint32_t serial) {
  return tls_->LaunchWithSharedWidget(app_id, inst_id, pid, serial);
}

int LauncherService::Impl::Launching(uint32_t serial) {
  return tls_->Launching(serial);
}

int LauncherService::Impl::LaunchCancel(uint32_t serial) {
  int ret = tls_->LaunchCancel(serial);
  done_ = true;
  return ret;
}

int LauncherService::Impl::LaunchDone(uint32_t serial) {
  int ret = tls_->LaunchDone(serial);
  done_ = true;
  return ret;
}

void LauncherService::Impl::OnRawBufferUpdated(
    std::shared_ptr<WlBuffer> wl_buffer,
    uint32_t time,
    uint32_t serial) {
}

void LauncherService::Impl::OnRawBufferMissing(uint32_t serial) {
}

void LauncherService::Impl::OnRawBufferChanged(uint32_t type,
    std::shared_ptr<WlBuffer> wl_buffer,
    int32_t fd,
    uint32_t size,
    uint32_t time,
    Direction direction,
    int32_t x,
    int32_t y,
    const std::string& shared_widget_info,
    uint32_t serial) {

  if (done_) {
    _W("Skip");
    return;
  }

  if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_TBM) {
    raw_image_ = std::shared_ptr<RawImage>(new RawImage(direction, x, y,
          shared_widget_info, wl_buffer));
    listener_->OnPrepareEvent(raw_image_, serial);
  } else if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {
    raw_image_ = std::shared_ptr<RawImage>(new RawImage(direction, x, y,
          shared_widget_info,
          std::shared_ptr<ImageFile>(new ImageFile(fd, size))));
    listener_->OnPrepareEvent(raw_image_, serial);
  } else {
    _W("Unknown type(%u)", type);
  }
}

void LauncherService::Impl::OnInputFilterChanged(uint32_t event_filter,
                                                 uint32_t serial) {
}

void LauncherService::Impl::OnPrepareEvent(uint32_t target_type,
    uint32_t target_res_id,
    const std::string& file_path,
    const std::string& file_group,
    Direction direction,
    int32_t x,
    int32_t y,
    const std::string& shared_widget_info,
    uint32_t serial) {
  if (target_type == TZSH_LAUNCHER_SERVICE_TARGET_TYPE_REMOTE_SURFACE) {
    done_ = false;
    listener_->OnResetEvent(serial);
    raw_image_.reset();
    trs_.reset(trs_mgr_->CreateTizenRemoteSurface(target_res_id,
          wl_surface_, direction, x, y, shared_widget_info, serial));
    trs_->AddListener(this);
    trs_->Redirect();
  } else if (target_type == TZSH_LAUNCHER_SERVICE_TARGET_TYPE_IMAGE) {
    raw_image_ = std::shared_ptr<RawImage>(new RawImage(direction, x, y,
          shared_widget_info, file_path));
    listener_->OnPrepareEvent(raw_image_, serial);
  } else if (target_type == TZSH_LAUNCHER_SERVICE_TARGET_TYPE_EDJE) {
    raw_image_ = std::shared_ptr<RawImage>(
        new RawImage(direction, x, y, shared_widget_info,
          file_path, file_group));
    listener_->OnPrepareEvent(raw_image_, serial);
  } else {
    _W("Unknown target type(%u)", target_type);
    listener_->OnErrorEvent(Error::WrongRequest, serial);
  }
}

void LauncherService::Impl::OnStopEvent(uint32_t serial) {
  listener_->OnStopEvent(serial);
  listener_->OnResetEvent(serial);
  raw_image_.reset();
  trs_.reset();
}

void LauncherService::Impl::OnErrorEvent(Error error, uint32_t serial) {
  if (error == Error::None)
    return;

  if (error == Error::Disqualified) {
    _E("Launcher has to clear LauncherService");
    listener_->OnErrorEvent(error, serial);
  } else {
    _W("Wrong request");
    listener_->OnErrorEvent(error, serial);
  }
}

void LauncherService::Impl::OnCleanupEvent(uint32_t serial) {
  listener_->OnCleanUpEvent(serial);
}

LauncherService::LauncherService(std::shared_ptr<Wl2Window> wl2_win,
                                 IEventListener* listener)
  : impl_(new Impl(this, std::move(wl2_win), listener)) {
}

LauncherService::~LauncherService() = default;

int LauncherService::Launch(const std::string& app_id,
                            const std::string& inst_id,
                            int pid,
                            uint32_t serial) {
  return impl_->Launch(app_id, inst_id, pid, serial);
}

int LauncherService::LaunchWithSharedWidget(const std::string& app_id,
                                            const std::string& inst_id,
                                            int pid,
                                            uint32_t serial) {
  return impl_->LaunchWithSharedWidget(app_id, inst_id, pid, serial);
}

int LauncherService::Launching(uint32_t serial) {
  return impl_->Launching(serial);
}

int LauncherService::LaunchCancel(uint32_t serial) {
  return impl_->LaunchCancel(serial);
}

int LauncherService::LaunchDone(uint32_t serial) {
  return impl_->LaunchDone(serial);
}

}  // namespace screen_connector
