/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Ecore_Wl2.h>
#include <tbm_surface.h>
#include <tbm_surface_internal.h>
#include <wayland-client.h>
#include <wayland-extension/tizen-extension-client-protocol.h>

#include "screen_connector_launcher_service/tizen_remote_surface_manager_internal.h"
#include "screen_connector_launcher_service/log_internal.h"

namespace screen_connector {

TizenRemoteSurfaceManager::TizenRemoteSurfaceManager() {
  if (!ecore_wl2_init()) {
    _E("ecore_wl2_init() is failed");
    return;
  }

  Ecore_Wl2_Display* wl2_dpy = ecore_wl2_connected_display_get(NULL);
  if (wl2_dpy == nullptr)
    wl2_dpy = ecore_wl2_display_connect(NULL);

  struct wl_display* wl_dpy = ecore_wl2_display_get(wl2_dpy);
  tbm_client_ = wayland_tbm_client_init(wl_dpy);
  tbm_ = wayland_tbm_client_get_wl_tbm(tbm_client_);

  struct wl_registry* registry = ecore_wl2_display_registry_get(wl2_dpy);
  Eina_Iterator* iter = ecore_wl2_display_globals_get(wl2_dpy);

  Ecore_Wl2_Global* global = nullptr;
  EINA_ITERATOR_FOREACH(iter, global) {
    if (!strcmp(global->interface, "tizen_remote_surface_manager")) {
      tzrs_mgr_ = static_cast<struct tizen_remote_surface_manager*>(
          wl_registry_bind(registry, global->id,
              &tizen_remote_surface_manager_interface, global->version));
      if (tzrs_mgr_)
        break;
    }
  }
  eina_iterator_free(iter);
}

TizenRemoteSurfaceManager::~TizenRemoteSurfaceManager() {
  if (tzrs_mgr_) {
    tizen_remote_surface_manager_destroy(tzrs_mgr_);
  }

  if (tbm_client_) {
    wayland_tbm_client_deinit(tbm_client_);
  }

  ecore_wl2_display_disconnect(ecore_wl2_connected_display_get(NULL));
  ecore_wl2_shutdown();
}

TizenRemoteSurface* TizenRemoteSurfaceManager::CreateTizenRemoteSurface(
    uint32_t res,
    std::shared_ptr<WlSurface> surface,
    Direction direction,
    int32_t x, int32_t y,
    std::string shared_widget_info,
    uint32_t serial) {
  struct tizen_remote_surface* tzrs =
    tizen_remote_surface_manager_create_surface_with_wl_surface(tzrs_mgr_,
        res, tbm_, surface->GetRaw());
  return new (std::nothrow) TizenRemoteSurface(tzrs, direction, x, y,
      std::move(shared_widget_info), serial);
}

}  // namespace screen_connector
