/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "screen_connector_common/util.h"
#include "screen_connector_launcher_service_evas/launcher_service_evas.h"
#include "screen_connector_launcher_service_evas/launcher_service_evas_implementation.h"
#include "screen_connector_launcher_service_evas/log_internal.h"

namespace screen_connector {

LauncherServiceEvas::Impl::Impl(LauncherServiceEvas* parent,
                                std::shared_ptr<EvasObject> win)
  : parent_(parent), win_(std::move(win)) {
}

const EvasObject& LauncherServiceEvas::Impl::GetWindow() {
  return *win_.get();
}

const Image& LauncherServiceEvas::Impl::GetImage() {
  return *image_.get();
}

void LauncherServiceEvas::Impl::Update(std::shared_ptr<RawImage> raw_image) {
  ImageType type = raw_image->GetImageType();
  if (type == ImageType::RemoteSurfaceTBM) {
    if (rs_img_.get() == nullptr) {
      _D("First added");
      rs_img_.reset(new Image(CreateEvasObjectImage()));
    }
    rs_img_->Update(raw_image);
    image_ = rs_img_;
  } else if (type == ImageType::RemoteSurfaceImageFile) {
    if (rs_img_.get() == nullptr) {
      _D("First added");
      rs_img_.reset(new Image(CreateEvasObjectImage()));
    }
    rs_img_->Update(raw_image);
    image_ = rs_img_;
  } else if (type == ImageType::SplashScreenImage) {
    if (ss_img_.get() == nullptr) {
      _D("First added");
      ss_img_.reset(new Image(CreateEvasObjectImage()));
    }
    ss_img_->Update(raw_image);
    image_ = ss_img_;
  } else if (type == ImageType::SplashScreenEdje) {
    if (ss_edje_.get() == nullptr) {
      _D("First added");
      ss_edje_.reset(new Image(CreateEdjeObject()));
    }
    ss_edje_->Update(raw_image);

    int width;
    int height;
    elm_win_screen_size_get(win_->GetRaw(), NULL, NULL, &width, &height);
    ss_edje_->SetSize(width, height);
    image_ = ss_edje_;
  }
}

void LauncherServiceEvas::Impl::Reset() {
  if (rs_img_.get() != nullptr) {
    _D("Reset raw image");
    rs_img_->Reset();
  }
}

Evas_Object* LauncherServiceEvas::Impl::CreateEvasObjectImage() {
  Evas_Object* raw = evas_object_image_filled_add(
    evas_object_evas_get(win_->GetRaw()));
  return raw;
}

Evas_Object* LauncherServiceEvas::Impl::CreateEdjeObject() {
  Evas_Object* raw = edje_object_add(evas_object_evas_get(win_->GetRaw()));
  return raw;
}

LauncherServiceEvas::LauncherServiceEvas(std::shared_ptr<EvasObject> win)
  : LauncherService(util::GetWl2Window(win->GetRaw()), this),
    impl_(new Impl(this, std::move(win))) {
}

LauncherServiceEvas::~LauncherServiceEvas() = default;

const EvasObject& LauncherServiceEvas::GetWindow() {
  return impl_->GetWindow();
}

const EvasObject& LauncherServiceEvas::GetImage() {
  return impl_->GetImage();
}

void LauncherServiceEvas::OnEvasPrepareEvent(const EvasObject& image,
    ImageType image_type,
    Direction direction,
    int32_t x,
    int32_t y,
    const std::string& shared_widget_info,
    uint32_t serial) {
}

void LauncherServiceEvas::OnEvasStopEvent(uint32_t serial) {
}

void LauncherServiceEvas::OnEvasErrorEvent(Error error,
                                           uint32_t serial) {
}

void LauncherServiceEvas::OnEvasCleanUpEvent(uint32_t serial) {
}

void LauncherServiceEvas::OnPrepareEvent(std::shared_ptr<RawImage> raw_image,
                                        uint32_t serial) {
  impl_->Update(raw_image);
  const Image& image = impl_->GetImage();
  OnEvasPrepareEvent(image, image.GetImageType(), image.GetDirection(),
      image.GetPositionX(), image.GetPositionY(),
      image.GetSharedWidgetInfo(), serial);
}

void LauncherServiceEvas::OnStopEvent(uint32_t serial) {
  OnEvasStopEvent(serial);
}

void LauncherServiceEvas::OnErrorEvent(Error error, uint32_t serial) {
  OnEvasErrorEvent(error, serial);
}

void LauncherServiceEvas::OnCleanUpEvent(uint32_t serial) {
  OnEvasCleanUpEvent(serial);
}

void LauncherServiceEvas::OnResetEvent(uint32_t serial) {
  impl_->Reset();
}

}  // namespace screen_connector
