/*
 * Copyright (c) 2016 - 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <sys/mman.h>
#include <dlog.h>
#include <Elementary.h>
#include <Ecore_Wl2.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>
#include <tbm_surface.h>
#include <tbm_surface_internal.h>
#include <uuid/uuid.h>
#include <wayland-client.h>
#include <wayland-tbm-client.h>

#include <map>
#include <list>
#include <algorithm>
#include <memory>

#include "screen_connector_toolkit_evas.h"
#include "screen_connector_remote_surface_evas/remote_surface_evas.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_WATCHER_EVAS"
#define SC_TOOLKIT_HANDLE_TAG "SC_TOOLKIT_HANDLE_TAG"

static Ecore_Event_Handler* __visibilityListener = nullptr;

class RemoteSurfaceEvasStub : public screen_connector::RemoteSurfaceEvas {
 public:
  RemoteSurfaceEvasStub(const std::string& id,
                        screen_connector::RemoteSurface::Type type,
                        screen_connector_toolkit_evas_ops* ops,
                        void* data,
                        std::shared_ptr<screen_connector::EvasObject> win)
      : screen_connector::RemoteSurfaceEvas(id, type, win),
        ops_(*ops), data_(data) {
  }

  RemoteSurfaceEvasStub(int rid, const std::string& id,
                        screen_connector::RemoteSurface::Type type,
                        screen_connector_toolkit_evas_ops* ops, void* data,
                        std::shared_ptr<screen_connector::EvasObject> win)
      : screen_connector::RemoteSurfaceEvas(rid, id, type, win),
        ops_(*ops), data_(data) {
  }

  virtual ~RemoteSurfaceEvasStub() = default;

  void OnEvasAdded(const std::string& appId, const std::string& instId,
                   int pid,
                   const screen_connector::EvasObject& image) override {
    if (ops_.added_cb)
      ops_.added_cb(appId.c_str(), instId.c_str(), pid, image.GetRaw(), data_);
  }

  void OnEvasRemoved(const std::string& appId, const std::string& instId,
                     int pid,
                     const screen_connector::EvasObject& image) override {
    if (ops_.removed_cb) {
      ops_.removed_cb(appId.c_str(), instId.c_str(), pid,
          image.GetRaw(), data_);
    }
  }

  void OnEvasChanged(const std::string& appId, const std::string& instId,
                     int pid,
                     const screen_connector::EvasObject& image) override {
    if (ops_.updated_cb) {
      ops_.updated_cb(appId.c_str(), instId.c_str(), pid,
          image.GetRaw(), data_);
    }
  }

 private:
  screen_connector_toolkit_evas_ops ops_;
  void* data_;
};

class Connector {
 public:
  explicit Connector(Evas_Object* win) : win_(win) { }
  ~Connector() { }

  void Add(RemoteSurfaceEvasStub* rs) {
    surfaces_.emplace_back(rs);
  }

  void Remove(RemoteSurfaceEvasStub* rs) {
    for (auto& i : surfaces_) {
      if (i.get() == rs) {
        surfaces_.remove(i);
        break;
      }
    }
  }

  void NotifyWindowVisibilityEvent(int type, void* event) {
    for (auto& rs : surfaces_) {
      rs->NotifyWindowVisibilityEvent(type, event);
    }
  }

  Evas_Object* GetWin() const { return win_; }

 private:
  Evas_Object *win_;
  std::list<std::unique_ptr<RemoteSurfaceEvasStub>> surfaces_;
};

static std::map<int, Connector*> __connectors;

extern "C" EXPORT_API int
screen_connector_toolkit_evas_init(Evas_Object *win,
                                   screen_connector_screen_type_e type) {
  if (__connectors.find(type) != __connectors.end())
    return -1;

  screen_connector::RemoteSurface::InitManager();
  __connectors[static_cast<int>(type)] = new Connector(win);
  return 0;
}

extern "C" EXPORT_API int screen_connector_toolkit_evas_fini(
    screen_connector_screen_type_e type) {
  auto i = __connectors.find(type);

  if (i == __connectors.end())
    return -1;

  __connectors.erase(i);

  if (__connectors.empty())
    screen_connector::RemoteSurface::FiniManager();

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_set_delayed_resuming_time(int ms) {
  screen_connector::RemoteSurface::SetDelayedResumingTime(ms);

  return 0;
}

extern "C" EXPORT_API screen_connector_toolkit_evas_h
screen_connector_toolkit_evas_add(screen_connector_toolkit_evas_ops* ops,
                                  char* id, screen_connector_screen_type_e type,
                                  void* data) {
  auto rs = new RemoteSurfaceEvasStub(id,
      (screen_connector::RemoteSurface::Type)type, ops, data,
      std::shared_ptr<screen_connector::EvasObject>(
      new screen_connector::EvasObject(
      __connectors[static_cast<int>(type)]->GetWin(), false)));

  __connectors[static_cast<int>(type)]->Add(rs);

  return static_cast<void*>(rs);
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_remove(screen_connector_toolkit_evas_h handle) {
  RemoteSurfaceEvasStub* rs = static_cast<RemoteSurfaceEvasStub*>(handle);

  for (int i = SCREEN_CONNECTOR_SCREEN_TYPE_WIDGET;
      i <= SCREEN_CONNECTOR_SCREEN_TYPE_ALL; i++) {
    auto sci = __connectors.find(i);

    if (sci == __connectors.end())
      continue;

    __connectors[i]->Remove(rs);
  }

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_start_visibility_notify(void) {
  if (__visibilityListener)
    return -1;

  __visibilityListener = ecore_event_handler_add(
      ECORE_WL2_EVENT_WINDOW_VISIBILITY_CHANGE, [](void* data, int type,
      void* event)->Eina_Bool {
    for (int i = SCREEN_CONNECTOR_SCREEN_TYPE_WIDGET;
         i <= SCREEN_CONNECTOR_SCREEN_TYPE_ALL; i++) {
      auto sci = __connectors.find(i);

      if (sci == __connectors.end())
        continue;

      __connectors[i]->NotifyWindowVisibilityEvent(type, event);
    }

    return ECORE_CALLBACK_RENEW;
  }, NULL);

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_stop_visibility_notify(void) {
  if (!__visibilityListener)
    return -1;

  ecore_event_handler_del(__visibilityListener);
  __visibilityListener = nullptr;

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_freeze_visibility(Evas_Object* obj,
                                                visibility_type type) {
  screen_connector::RemoteSurface* rs =
      static_cast<screen_connector::RemoteSurface*>(evas_object_data_get(obj,
      SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  switch (type) {
    case VISIBILITY_TYPE_FULLY_OBSCURED:
      rs->Freeze(false);
      break;
    case VISIBILITY_TYPE_UNOBSCURED:
    case VISIBILITY_TYPE_PARTIALLY_OBSCURED:
      rs->Freeze(true);
      break;
    default:
      return -1;
  }

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_thaw_visibility(Evas_Object* obj) {
  screen_connector::RemoteSurface* rs =
      static_cast<screen_connector::RemoteSurface*>(evas_object_data_get(obj,
      SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  rs->Thaw();

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_send_visibility(Evas_Object* obj,
                                              visibility_type type) {
  screen_connector::RemoteSurface* rs =
      static_cast<screen_connector::RemoteSurface*>(evas_object_data_get(obj,
      SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  switch (type) {
    case VISIBILITY_TYPE_FULLY_OBSCURED:
      rs->SendVisibility(false);
      break;
    case VISIBILITY_TYPE_UNOBSCURED:
    case VISIBILITY_TYPE_PARTIALLY_OBSCURED:
      rs->SendVisibility(true);
      break;
    default:
      return -1;
  }

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_send_mouse_up(Evas_Object* obj) {
  screen_connector::RemoteSurfaceEvas* rs =
      static_cast<screen_connector::RemoteSurfaceEvas*>(
      evas_object_data_get(obj, SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  rs->SendMouseUp();

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_send_touch_cancel(Evas_Object* obj) {
  screen_connector::RemoteSurface* rs =
      static_cast<screen_connector::RemoteSurface*>(evas_object_data_get(obj,
      SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  rs->SendTouchCancel();

  return 0;
}

extern "C" EXPORT_API screen_connector_toolkit_evas_h
screen_connector_toolkit_evas_add_by_rid(screen_connector_toolkit_evas_ops* ops,
                                         int res_id,
                                         screen_connector_screen_type_e type,
                                         void* data) {
  char id_str[70];
  char uuid[37];
  uuid_t u;

  uuid_generate(u);
  uuid_unparse(u, uuid);
  snprintf(id_str, sizeof(id_str), "%s:%d", uuid, res_id);

  auto rs = new RemoteSurfaceEvasStub(res_id, id_str,
      (screen_connector::RemoteSurface::Type)type, ops, data,
      std::shared_ptr<screen_connector::EvasObject>(
      new screen_connector::EvasObject(
      __connectors[static_cast<int>(type)]->GetWin(), false)));

  __connectors[static_cast<int>(type)]->Add(rs);

  return static_cast<void*>(rs);
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_get_rid(Evas_Object* obj, int* resource_id) {
  screen_connector::RemoteSurface* rs =
      static_cast<screen_connector::RemoteSurface*>(evas_object_data_get(obj,
      SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  *resource_id = rs->GetRid();
  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_get_pid(Evas_Object* obj, int *pid) {
  screen_connector::RemoteSurface* rs =
      static_cast<screen_connector::RemoteSurface*>(evas_object_data_get(obj,
      SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  *pid = rs->GetPid();
  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_is_visible(Evas_Object* obj, bool* is_visible) {
  screen_connector::RemoteSurfaceEvas* rs =
      static_cast<screen_connector::RemoteSurfaceEvas*>(
      evas_object_data_get(obj, SC_TOOLKIT_HANDLE_TAG));

  if (!rs)
    return -1;

  if (rs->IsVisibleArea())
    *is_visible = true;
  else
    *is_visible = false;

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_bind(screen_connector_toolkit_evas_h h,
                                   Evas_Object* win) {
  screen_connector::RemoteSurfaceEvas* rs =
      static_cast<screen_connector::RemoteSurfaceEvas*>(h);

  if (!rs)
    return -1;

  LOGI("bind window");
  rs->Bind(screen_connector::EvasObject(win, false));

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_unbind(screen_connector_toolkit_evas_h h) {
  screen_connector::RemoteSurfaceEvas* rs =
      static_cast<screen_connector::RemoteSurfaceEvas*>(h);

  if (!rs)
    return -1;

  LOGI("bind window");
  rs->Unbind();

  return 0;
}

extern "C" EXPORT_API visibility_type
screen_connector_toolkit_evas_auto_visibility(
    screen_connector_toolkit_evas_h h) {
  screen_connector::RemoteSurfaceEvas* rs =
      static_cast<screen_connector::RemoteSurfaceEvas*>(h);

  if (!rs)
    return VISIBILITY_TYPE_UNKNOWN;

  return (visibility_type)rs->GetVisibility();
}

extern "C" EXPORT_API int
screen_connector_toolkit_evas_set_changed_event_filter(
    screen_connector_toolkit_evas_h toolkit_evas_h,
    screen_connector_evas_changed_event_filter_type type) {
  screen_connector::RemoteSurfaceEvas* rs =
      static_cast<screen_connector::RemoteSurfaceEvas*>(toolkit_evas_h);

  if (!rs)
    return -1;

  rs->SetChangedEventFilter(
      static_cast<screen_connector::RemoteSurface::ChangedEventFilter>(type));

  return 0;
}

